/**
 * @file      canvas_jpegio.h
 * @brief     JPEG IO support for canvas.
 * @date      2010-12-09
 *
 * @copyright
 * Copyright 2010 Japan Aerospace Exploration Agency
 *
 */

#ifndef _CANVAS_JPEGIO_H
#define _CANVAS_JPEGIO_H

#ifdef HAVE_JPEGLIB_H

#include <jpeglib.h>
#include <jerror.h>
#include "canvas.h"

typedef int (*canvasJpegioAfterCreateCallback)(canvas *cv,
                                               struct jpeg_compress_struct *cinfo,
                                               void *userdata);

/*
 * JPEG 入出力フラグ。
 */
#define CVFormatJpegProgressive  0x01000000 /** Progressive JPEG を作成 */

__BEGIN_DECLS
/**
 * コンストラクタ：JPEG 読み込み
 * attr はマスク値 (現在は使われていない) です。
 */
canvas *cv_load_jpeg(const char *path, int attr);

/**
 * canvas データを JPEG 形式で保存します。
 * @param self
 *          canvas オブジェクト。
 * @param path
 *          出力ファイルパス。
 * @param quality
 *          品質。0(最低) - 100(最高)。数値が低いほど圧縮率は高くなります。
 * @param attr
 *          マスク。
 *          * CVFormatJpegProgressive: Progressive JPEG を作成します。
 * @param cb
 *          画像データが書きだされる前に呼ばれるコールバック関数。
 * @param userdata
 *          コールバック関数の引数。
 * @result 成功したら 1、失敗したら 0。
 * @see cv_write_jpeg()
 */
int cv_save_jpeg(canvas *self, const char *path, int quality, int attr,
                 canvasJpegioAfterCreateCallback cb, void *userdata);

/**
 * canvas データを JPEG 形式で保存します。
 * @param self      canvas instance.
 * @param io        出力ストリーム。wb で開いている必要があります。
 * @param quality   品質。0(最低) - 100(最高)。数値が低いほど圧縮率は高く
 *                  なります。
 * @param attr      出力ファイルフォーマットを制御するマスク。
 *                  * CVFormatJpegProgressive: Progressive JPEG を作成します。
 * @param cb        画像データが書きだされる前に呼ばれるコールバック関数。
 * @param userdata  コールバック関数の引数。
 * @result success: 1. error: 0.
 * @see cv_write_jpeg()
 */
int cv_save_jpeg_with_std(canvas *self, FILE *io, int quality, int attr,
                          canvasJpegioAfterCreateCallback cb, void *userdata);

/**
 * canvas データを jpeg_compress_struct オブジェクトに書きだします。
 * @param self
 *          canvas オブジェクト。
 * @param cinfo
 *          jpeg_compress_struct オブジェクト。
 * @param quality
 *          品質。0(最低) - 100(最高)。数値が低いほど圧縮率は高くなります。
 * @param attr
 *          マスク。
 *          * CVFormatJpegProgressive: Progressive JPEG を作成します。
 * @param cb
 *          画像データが書きだされる前に呼ばれるコールバック関数。
 * @param userdata
 *          コールバック関数の引数。
 * @result 成功したら 1、失敗したら 0。
 * @see cv_save_jpeg()
 */
int cv_write_jpeg(canvas *self, struct jpeg_compress_struct *cinfo, int quality,
                  int attr, canvasJpegioAfterCreateCallback cb, void *userdata);

__END_DECLS

#endif // HAVE_JPEGLIB_H

#endif // _CANVAS_JPEGIO_H
