/**
 * @file      flow_common.h
 * @brief     flow_se and flow_ig common codes
 * @date      2011-03-11
 *
 * @copyright
 * Copyright 2011 Japan Aerospace Exploration Agency
 *
 */

#ifndef _FLOW_COMMON_H
#define _FLOW_COMMON_H

#include <SpiceUsr.h>

//
// defined in main
//
extern char app_path[], app_name[];

#ifndef M_PI
#define M_PI 3.14159265358979323846
#endif

#define rad2deg(x)    ((x)*180.0/M_PI)
#define deg2rad(x)    ((x)*M_PI/180.0)
#define max(a,b)      (((a) < (b)) ? (b) : (a))
#define min(a,b)      (((a) < (b)) ? (a) : (b))

//
// object IDs defined in:
//   http://naif.jpl.nasa.gov/pub/naif/toolkit_docs/C/req/naif_ids.html
//
#define OBJECT_ID_SUN           10
#define OBJECT_ID_MOON          301
#define OBJECT_ID_MERCURY       199
#define OBJECT_ID_VENUS         299
#define OBJECT_ID_EARTH         399
#define OBJECT_ID_MARS          499
#define OBJECT_ID_JUPITER       599
#define OBJECT_ID_SATURN        699
#define OBJECT_ID_URANUS        799
#define OBJECT_ID_NEPTUNE       899
#define OBJECT_ID_PLUTO         999

//
// object types
//
#define OBJECT_TYPE_INVALID                     0
#define OBJECT_TYPE_ASAT                        1 // artificial sat. (earth orbitting spacecraft)
#define OBJECT_TYPE_SPACECRAFT          2 // other spacecraft or it's instrument
#define OBJECT_TYPE_BARYCENTER          3 // system barycenter
#define OBJECT_TYPE_SUN                         4
#define OBJECT_TYPE_PLANET                      5 // including pluto
#define OBJECT_TYPE_SATELLITE           6
#define OBJECT_TYPE_SMALL_BODY          7 // comet, asteroid or other object

enum FLOWImageFormat {
  FLOWImageFormatAuto,
  FLOWImageFormatFITS,
  FLOWImageFormatJPEG,
  FLOWImageFormatPNG,
  FLOWImageFormatUnkown,
  FLOWImageFormatError
};

enum FLOWProjectionMode {
  FLOWProjectionModeFill = 1,               /** 縦、横が画面に収まるようにスケール。*/
  FLOWProjectionModeScaleAspectFit,         /** カメラの比率を保ったまま全体が収まるようにスケール。*/
  FLOWProjectionModeScaleAspectFill,        /** カメラの比率を保ったまま画面全体に表示されるようにスケール。*/
};

#define FLOW_CONFIG_DIR                 ".flow"

#ifdef ENABLE_WMS
#define FLOW_WMS_CACHE_FILE             "/wms_cache.edb"
#endif

__BEGIN_DECLS
double vec2d_cross_product(const double *v1, const double *v2);
double vec2d_dot_product(const double *v1, const double *v2);
void vec_middle(const SpiceDouble *vec1, SpiceDouble *vec2, SpiceDouble *dst);
void vec_padist(const SpiceDouble *vec1, const SpiceDouble *vec2, SpiceDouble *_pa, SpiceDouble *_dist);
void vec_plotxy(const SpiceDouble *bv, const SpiceDouble *ltv, const SpiceDouble *tv, double r, double *_x, double *_y);
double *bounds_top_left(int nbounds, double *bounds);
double *bounds_bottom_right(int nbounds, double *bounds);
int get_object_type(int object_id);
void mpc_get_bounds_rect(const char *shape, const double *boresight, int nbounds, const double *bounds, double *rect);
void mpc_get_projection_rect(enum FLOWProjectionMode mode, double w, double h, double left, double top, double right, double bottom, double *rect);
void mpc_get_boresight_rotation_matrix(SpiceDouble boresight[3], SpiceDouble r[3][3]);
void mpc_rotate_frame(SV_FRAME* frame);
/**
 * config directory 下にある path の絶対パスを作成して返します。
 * buf は少なくとも nbuf バイト確保しておくか、NULL にします。
 * NULL の場合は内部で malloc した結果が返ります。
 * 作成されたパスが nbuf を超える場合は NULL が返ります。
 * @param path  設定ディレクトリ以下のパス。
 * @param buf   絶対パスが代入される。
 * @param nbuf  buf の確保サイズ。
 * @param nlen  実際に必要なサイズ。
 * @result  buf が NULL か十分確保されている場合は作成結果のパス。
 *          そうでない場合は NULL。
 */
char *mpc_get_config_path(const char *path, char *buf, size_t nbuf, size_t *nlen);

/**
 * 設定ディレクトリを作成します。
 * 成功すると 0 を返します。
 */
int mpc_create_config_dir(void);

/**
 * 親ディレクトリが存在しない場合はさかのぼってディレクトリを作成する mkdir (2)。
 * @param path ディレクトリを作成する path。
 * @param mode create mode。
 * @result 成功したら 0。失敗したら最後のコマンドの戻り値。
 */
int mpc_mkdir_p(const char *path, int mode);

/**
 * 角度を max-360.0 <= x <= max に正規化します。
 * @param deg  角度 (degree)。
 * @param max  角度の最大値 (degree)。
 * @result 正規化後の角度。
 */
double mpc_degnorm(double deg, double max);

/**
 * 拡張子から画像フォーマットを判定します。
 * @param path ファイルパス。
 */
enum FLOWImageFormat predict_image_format_from_filename(const char *path);

/**
 * 実行ファイル名およびパス名を取得
 * @param arg0
 * @param name
 * @param path
 */
void get_appname(char *arg0, char *name, char *path);
__END_DECLS

#endif // _FLOW_COMMON_H
