/**
 * @file      flowse_lua_bind.c
 * @brief     SPICE/FLOW lua binding
 * @date      2010-03-12
 *
 * @copyright
 * Copyright 2010 Japan Aerospace Exploration Agency
 *
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <assert.h>
#include "SpiceUsr.h"
#include "flowse_lua_bind.h"

static lua_State *Lua = NULL;
static const char *lua_app_name = NULL;

#if LUA_VERSION_NUM>=502
#define lua_objlen lua_rawlen
#endif

//////////////////////////////////////////////////////////////////////
// init / end
//////////////////////////////////////////////////////////////////////

/**
 * LUA ステートを作成し、flow 用関数をバインドします。
 * @result LUA ステート。失敗したら NULL。
 */
int flowse_lua_init(const char* app_name)
{
  lua_State *lua;

  // 既に初期化済み
  if (Lua) {
    return (int)Lua;
  }
  lua_app_name = app_name;

  // LUA の初期化
  lua = luaL_newstate();
  luaL_openlibs(lua);           // 標準ライブラリを読み込む

  // LUA 関数を設定する
  spice_lua_init(lua);

  Lua = lua;

  return (int)lua;
}

/*
 * 一時ファイルを作成して、そのファイルを LUA として読み込む。
 */
int flowse_lua_parse_fd(int infd)
{
  char luafile[] = "flow_se_conf_XXXXXX";
  int fd;
  char buf[512];
  char msg[256];
  int rc;
  int wc;

  if ((fd = mkstemp(luafile)) == -1) {
    sprintf(msg, "%s: cannot create temporary file", lua_app_name);
    perror(msg);
    return 0;
  }
  while ((rc = read(infd, buf, sizeof(buf))) > 0) {
    if ((wc = write(fd, buf, rc)) < 0) {
      close(fd);
      unlink(luafile);
      sprintf(msg, "%s: cannot write to temporary file", lua_app_name);
      perror(msg);
      return 0;
    }
  }
  close(fd);
  if (rc < 0) {
    unlink(luafile);
    sprintf(msg, "%s: cannot read temporary file", lua_app_name);
    perror(msg);
    return 0;
  }
  if (!flowse_lua_parse(luafile)) {
    unlink(luafile);
    return 0;
  }
  unlink(luafile);
  return 1;
}

/**
 * LUA フォーマットの設定ファイルを読み込み、実行します。
 * @param path  設定ファイルのパス。
 * @result 成功したら 1、失敗したら 0。
 */
int flowse_lua_parse(const char *path)
{
  lua_State *lua = flowse_lua_get_state();
  int err;
  const char *errmsg;

  luaL_loadfile(lua, path);
  if ((err = lua_pcall(lua, 0, 0, 0)) != 0) {
    errmsg = lua_tostring(lua, 1);
    fprintf(stderr, "%s: (config file error): %s\n", lua_app_name, errmsg);
    return 0;
  }
  return 1;
}

/**
 * LUA ステートを開放します。
 */
void flowse_lua_end(void)
{
  if (Lua) {
    lua_close(Lua);
  }
  Lua = NULL;
}

/**
 * LUA ステートを取得します。
 * @result lua_State 変数。
 */
lua_State *flowse_lua_get_state(void)
{
  return Lua;
}

/**
 * グローバル変数 name が定義されているか調べます。
 * @param name  変数名。
 */
int flowse_lua_is_defined(const char *name)
{
  int is_defined = 1;
  lua_State *lua = flowse_lua_get_state();
  if (!lua_checkstack(lua, 1)) {
    luaL_error(lua, "cannot take stack space %s", name);
    return 0;             /* スタックが伸ばせない */
  }
  lua_getglobal(lua, name);       /* 値がなくてもスタックは増える */
  if (lua_isnil(lua, -1)) {
    is_defined = 0;             /* スタックトップが nil = 値がない */
  }
  lua_pop(lua, 1);
  return is_defined;
}

/**
 * グローバル変数 name が定義されているか調べ、その型を返します。
 * @param name  変数名。
 */
int flowse_lua_get_type(const char *name)
{
  lua_State *lua = flowse_lua_get_state();
  int type;

  if (!lua_checkstack(lua, 1)) {
    luaL_error(lua, "cannot take stack space %s", name);
    return 0;             /* スタックが伸ばせない */
  }
  lua_getglobal(lua, name);
  type = lua_type(lua, -1);
  lua_pop(lua, 1);

  return type;
}

/**
 * int 型のグローバル変数 name の値を返します。
 */
int flowse_lua_get_integer(const char *name)
{
  lua_State *lua = flowse_lua_get_state();
  int value;
  int level;

  if (LUA_TNUMBER != flowse_lua_get_type(name)) {
    luaL_error(lua, "value %s is not number", name);
    return 0.0;
  }
  if (!lua_checkstack(lua, 1)) {
    luaL_error(lua, "cannot take stack space %s", name);
    return 0.0;             /* スタックが伸ばせない */
  }
  level = lua_gettop(lua);
  lua_getglobal(lua, name);
  if (level >= lua_gettop(lua)) {
    luaL_error(lua, "cannot get value %s", name);
    return 0.0;             /* スタックが積まれていない */
  }
  value = lua_tointeger(lua, -1);
  lua_pop(lua, 1);

  return value;
}

/**
 * Double 型のグローバル変数 name の値を返します。
 */
double flowse_lua_get_double(const char *name)
{
  lua_State *lua = flowse_lua_get_state();
  double value;
  int level;

  if (LUA_TNUMBER != flowse_lua_get_type(name)) {
    luaL_error(lua, "value %s is not number", name);
    return 0.0;
  }
  if (!lua_checkstack(lua, 1)) {
    luaL_error(lua, "cannot take stack space %s", name);
    return 0.0;             /* スタックが伸ばせない */
  }
  level = lua_gettop(lua);
  lua_getglobal(lua, name);
  if (level >= lua_gettop(lua)) {
    luaL_error(lua, "cannot get value %s", name);
    return 0.0;             /* スタックが積まれていない */
  }
  value = lua_tonumber(lua, -1);
  lua_pop(lua, 1);

  return value;
}

/**
 * LUA のグローバル変数から文字列を取得し、alloc して返します。
 * 返り値は free でメモリを開放してください。
 * @param name  グローバル変数名。
 * @result 変数の値。値が nil、変換不可能、または存在しない場合は NULL。
 */
char *flowse_lua_copy_string(const char *name)
{
  lua_State *lua = flowse_lua_get_state();
  const char *str;
  char *ptr;
  int level;

  if (!lua_checkstack(lua, 1)) {
    luaL_error(lua, "cannot take stack space %s", name);
    return NULL;             /* スタックが伸ばせない */
  }
  level = lua_gettop(lua);
  lua_getglobal(lua, name);
  if (level >= lua_gettop(lua)) {
    luaL_error(lua, "cannot get value %s", name);
    return NULL;             /* スタックが積まれていない */
  }
  str = lua_tostring(lua, -1);
  if (!str) {
    luaL_error(lua, "cannot convert %s to string", name);
    lua_pop(lua, 1);             /* 文字列変換失敗 */
    return NULL;
  }
  ptr = strdup(str);
  lua_pop(lua, 1);

  return ptr;
}

/**
 * LUA のグローバル変数の配列長を返します。
 * @param name  LUA の変数名。
 * @result 配列の長さ。配列でないか、値が定義されていない場合は (size_t)-1。
 */
size_t flowse_lua_get_array_length(const char *name)
{
  lua_State *lua = flowse_lua_get_state();
  size_t length;

  if (!lua_checkstack(lua, 1)) {
    luaL_error(lua, "cannot take stack space %s", name);
    return (size_t)-1;             /* スタックが伸ばせない */
  }
  if (LUA_TTABLE != flowse_lua_get_type(name)) {
    return (size_t)-1;             /* table ではない */
  }
  lua_getglobal(lua, name);
  length = lua_objlen(lua, -1);
  lua_pop(lua, 1);

  return length;
}

/**
 * LUA のグローバル変数から配列の要素 (文字列) を取得し、alloc して返します。
 * 返り値は free でメモリを開放してください。
 * @param name  LUA の変数名。
 * @param index 要素のインデックス。LUA の規則とは異なり、先頭の要素インデッ
 *              クスは 0 です。
 * @result table の要素文字列。スタックが伸ばせない、変数が存在しない、イン
 *               デックスが範囲外、または要素が文字列に変換できない場合は
 *               NULL。
 */
char *flowse_lua_copy_array_item_as_string(const char *name, int index)
{
  lua_State *lua = flowse_lua_get_state();
  size_t length;
  const char *str;
  char *ptr;

  if (!lua_checkstack(lua, 2)) {
    luaL_error(lua, "cannot take stack space %s", name);
    return NULL;             /* スタックが伸ばせない */
  }
  length = flowse_lua_get_array_length(name);
  if (length == (size_t)-1) {
    return NULL;
  }
  if (index < 0 || length <= index) {
    return NULL;             /* index が範囲外 */
  }
  lua_getglobal(lua, name);
  lua_pushnumber(lua, (lua_Number)index+1);       /* index -> top */
  lua_gettable(lua, -2);         /* name[index] -> top (index is removed) */
  str = lua_tostring(lua, -1);
  if (!str) {
    /* 文字列に変換できない */
    lua_pop(lua, 2);
    return NULL;
  }
  ptr = strdup(str);
  lua_pop(lua, 2);

  return ptr;
}

/**
 * object_table の指定オブジェクトの定義済み配列要素数を得る。
 * @param name   LUA の object_table 変数名。
 * @param object 天体ID（object_table 配列の添え字）
 * @result 配列の長さ。配列でないか、値が定義されていない場合は (size_t)-1。
 */
size_t flowse_lua_object_table_get_length(const char *name, const char *object)
{
  lua_State *lua = flowse_lua_get_state();
  size_t length;

  if (!lua_checkstack(lua, 2)) {
    luaL_error(lua, "cannot take stack space %s", name);
    return (size_t)-1;             /* スタックが伸ばせない */
  }
  if (LUA_TTABLE != flowse_lua_get_type(name)) {
    return (size_t)-1;             /* table ではない */
  }
  lua_getglobal(lua, name);
  lua_pushstring(lua, object);
  lua_gettable(lua, -2);         /* name[object] -> top (index is removed) */
  if (lua_isnil(lua, -1)) {
    lua_pop(lua, 2);
    return (size_t)-1;             /* name[object] が定義されていない */
  }
  length = lua_objlen(lua, -1);
  lua_pop(lua, 2);

  return length;
}

/**
 * object_table の指定オブジェクト/インデックスの値を取得する。
 * @param name   LUA の object_table 変数名。
 * @param object 天体ID（object_table 配列の添え字）
 * @param index  オブジェクト定義配列添え字
 * @param tag    取得すべきプロパティ名
 * @result オブジェクト定義値。値が定義されていない場合は NULL。
 */
char *flowse_lua_object_table_copy_value(const char *name, const char *object,
                                         int index, const char *tag)
{
  lua_State *lua = flowse_lua_get_state();
  size_t length;
  const char *str;
  char *ptr;

  if (!lua_checkstack(lua, 4)) {
    luaL_error(lua, "cannot take stack space %s", name);
    return NULL;             /* スタックが伸ばせない */
  }
  if (LUA_TTABLE != flowse_lua_get_type(name)) {
    return NULL;             /* table ではない */
  }
  lua_getglobal(lua, name);
  lua_pushstring(lua, object);
  lua_gettable(lua, -2);         /* name[object] -> top (object is removed) */
  if (lua_isnil(lua, -1)) {
    lua_pop(lua, 2);
    return NULL;             /* name[object] が定義されていない */
  }
  length = lua_objlen(lua, -1);
  if (index < 0 || length <= index) {
    lua_pop(lua, 2);
    return NULL;             /* index が範囲外 */
  }
  lua_pushnumber(lua, (lua_Number)index+1);       /* index -> top */
  lua_gettable(lua, -2);         /* name[object][index] -> top (index is removed) */
  if (lua_isnil(lua, -1)) {
    lua_pop(lua, 3);
    return NULL;             /* name[object][index] が定義されていない */
  }
  lua_pushstring(lua, tag);
  lua_gettable(lua, -2);         /* name[object][index][tag] -> top (tag is removed) */
  if (lua_isnil(lua, -1)) {
    lua_pop(lua, 4);
    return NULL;             /* name[object][index][tag] が定義されていない */
  }
  str = lua_tostring(lua, -1);
  if (!str) {
    /* 文字列に変換できない */
    lua_pop(lua, 4);
    return NULL;
  }
  ptr = strdup(str);
  lua_pop(lua, 4);

  return ptr;
}
