/**
 * @file      poly_mesh.h
 * @brief     3-Dimentional polygon mesh
 * @date      2010-04-02
 *
 * @copyright
 * Copyright 2010 AstroArts Inc.
 *
 */

#ifndef _POLY_MESH_H
#define _POLY_MESH_H

typedef struct _poly_mesh {
  size_t polygon_size;
  size_t vertex_size;
  double *vertex;
  double *texcoord;
  int    *index;
} poly_mesh;

typedef enum _PolyMeshType {
  PolyMeshTypeSphere = 0
} PolyMeshType;

typedef enum _PolyMeshAccuracy {
  PolyMeshAccuracyNormal = 0,
  PolyMeshAccuracyMedium,
  PolyMeshAccuracyHigh
} PolyMeshAccuracy;

__BEGIN_DECLS

/**
 * モデルを生成します。
 */
poly_mesh *mesh_alloc(PolyMeshType type, PolyMeshAccuracy accuracy);

/**
 * 球体モデルを生成します。
 * @param xsplit  x (longitude) 方向分割数。
 * @param ysplit  y (latitude) 方向分割数。
 */
poly_mesh *mesh_alloc_sphere(int xsplit, int ysplit);

/**
 * 球体の部分モデルを生成します。
 * minlat < maxlat, minlng < maxlng でない場合はエラーになります。
 * @param xsplit  x (longitude) 方向分割数。
 * @param ysplit  y (latitude) 方向分割数。
 * @param minlat  緯度の最小値 (-90 <= deg < 90).
 * @param minlng  経度の最小値 (0 <= deg < 360).
 * @param maxlat  緯度の最大値 (-90 < deg <= 90).
 * @param maxlng  経度の最大値 (0 < deg <= 360).
 */
poly_mesh *mesh_alloc_sphere_part(int xsplit, int ysplit, double minlat, double minlng, double maxlat, double maxlng);

/**
 * モデルを移動します。
 */
void mesh_move(poly_mesh *self, double x, double y, double z);

/**
 * モデルを拡大・縮小します。
 */
void mesh_scale(poly_mesh *self, double x, double y, double z);

/**
 * モデルを回転します。
 */
void mesh_rotate(poly_mesh *self, double mtx[3][3]);

/**
 * モデルを開放します。
 */
void mesh_free(poly_mesh *self);

__END_DECLS

#endif /* _POLY_MESH_H */
