/**
 * @file      zbuf.h
 * @brief     Z-buffer for 3D rendering
 * @date      2011-03-11
 *
 * @copyright
 * Copyright 2010-2011 Japan Aerospace Exploration Agency
 *
 */

#ifndef _ZBUF_H_
#define _ZBUF_H_

#include <stdint.h>

typedef int ZBUFBool;
#define ZBUF_TRUE   1
#define ZBUF_FALSE  0

enum ZBUFCondition {
  ZBUFConditionLess,
  ZBUFConditionEqualLess,
  ZBUFConditionGreater,
  ZBUFConditionEqualGreater,
  ZBUFConditionEqual,
  ZBUFConditionNotEqual,
};

typedef struct _zbuf {
  double *buffer;
  size_t width;
  size_t height;
  enum ZBUFCondition condition;
  double near;
  double far;
} zbuf;

__BEGIN_DECLS

/**
 * コンストラクタ。
 */
zbuf *zbuf_create(int width, int height, enum ZBUFCondition cond, double near, double far);

/**
 * デストラクタ。
 */
void zbuf_free(zbuf *self);

/**
 * Z 値をクリアし、全て far にセットします。
 */
void zbuf_clear(zbuf *self);

/**
 * near, far を再設定します。
 */
void zbuf_set_range(zbuf *self, double near, double far);

/**
 * 座標の Z 値を取得します。
 * @param x     座標 (x)。
 * @param y     座標 (y)。
 */
double zbuf_z(zbuf *self, int x, int y);

/**
 * 座標の Z 値が指定条件に合うかどうか調べます。
 * @param self  zbuf instance.
 * @param x     座標 (x)。
 * @param y     座標 (y)。
 * @param z     比較する Z 値。
 * @result  条件に合致したかどうか。
 */
ZBUFBool zbuf_test(zbuf *self, int x, int y, double z);

/**
 * 座標の Z 値が指定条件に合うかどうか調べ、マッチしたら Z 値を上書き
 * します。
 * @param self  zbuf instance.
 * @param x     座標 (x)。
 * @param y     座標 (y)。
 * @param z     比較する Z 値。
 * @result  条件に合致したかどうか。
 */
ZBUFBool zbuf_set(zbuf *self, int x, int y, double z);

__END_DECLS

#endif /* _ZBUF_H_ */
